#!/usr/bin/python
#
# PE326 Assignment
# 
# Infinite Quantum Well of width 2*a with a potential step of width 2*w in the
# center.
# 
# This file solely drives the GUI. The bulk of the Computation is done in
# PE326Computer.py.
# 
# The author is not willing to confess authorship to this mess.
# 

import sys, os, os.path
import math
import wx, wx.html, wx.lib.wxpTag, wx.lib.filebrowsebutton
import Image, ImageDraw, ImageFont
import PngImagePlugin
import PE326Computer

global use_eV

eV = 1.60211773e-19

def eVToJ(s):
	ans = None
	try:
		ans = float(s)
	except:
		return None
	return ans*eV

def JToeV(s):
	ans = None
	try:
		ans = float(s)
	except:
		return None
	return ans/eV

class MiniGraph(wx.Panel):
	def __init__(self, parent):
		wx.Panel.__init__(self, parent, -1, style=wx.FULL_REPAINT_ON_RESIZE)
		self.SetBackgroundColour(wx.WHITE)
		self.Bind(wx.EVT_PAINT, self.OnPaint)
		self.Bind(wx.EVT_SIZE, self.Resize)
		self.a = self.w = self.v = self.start = self.end = 0
		self.sane = False
		self.insane_string = None

		self.margin_left = 5
		self.margin_right = 5
		self.margin_top = 5
		self.margin_bottom = 5

		self.ymax = 100
		self.ymin = 0
		
	def Resize(self, evt):
		self.SetSize(self.GetMaxSize())

	def ptToX(self, x):
		viewport_w = self.width - self.margin_left - self.margin_right
		return int(self.margin_left + ((x - self.start) * viewport_w) / (self.end - self.start))

	def ptToY(self, y):
		viewport_h = self.height - self.margin_top - self.margin_bottom
		return int(self.margin_top + viewport_h - (((y - self.ymin) * viewport_h) / (self.ymax - self.ymin)))

	def SetSanity(self, sane, str = None):
		self.sane = sane
		self.insane_string = str

	def UpdateValues(self, a, w, v):
		self.a = a
		self.w = w
		self.v = v

		self.start = -self.a
		self.end = self.a

	def Plot(self, dc):
		self.width, self.height = self.GetSizeTuple()
		a, w, v = self.a, self.w, self.v
		if v == 0 or w == 0:
			v = 0
			w = 0
		dc.BeginDrawing()

		if not self.sane:
			dc.SetTextForeground('RED')
			dc.DrawText('Invalid parameters!', 10, 10)
			if self.insane_string != None:
				dc.DrawText(self.insane_string, 10, 30)
			return

		# Box
		dc.SetPen(wx.Pen('BLACK', 2, wx.SOLID))
		vh = 100.0 / (5*eV) * v
		dc.DrawLine(self.ptToX(-a), self.ptToY(0), self.ptToX(-a), self.ptToY(150))
		dc.DrawLine(self.ptToX(-a), self.ptToY(0), self.ptToX(-w), self.ptToY(0))
		dc.DrawLine(self.ptToX(-w), self.ptToY(0), self.ptToX(-w), self.ptToY(vh))
		dc.DrawLine(self.ptToX(-w), self.ptToY(vh), self.ptToX(w), self.ptToY(vh))
		dc.DrawLine(self.ptToX(w), self.ptToY(0), self.ptToX(w), self.ptToY(vh))
		dc.DrawLine(self.ptToX(w), self.ptToY(0), self.ptToX(a), self.ptToY(0))
		dc.DrawLine(self.ptToX(a), self.ptToY(0), self.ptToX(a), self.ptToY(150))

		dc.EndDrawing()

	def OnPaint(self, evt):
		dc = wx.PaintDC(self)
		dc.Clear()
		self.Plot(dc)

class AboutBox(wx.Dialog):
	text = '''
<html>
<body bgcolor="#ccddbb">
<center>
<h1 align="center">Physical Electronics 326<br>Assignment</h1>

<p>
<font size="-1">
This demonstration was brought to you by the following individuals,
some of who have far too much time on their hands...</font>
</p>

<p>
<i>Bernard Blackham<br>
Alison Steer<br>
Mark Rankilor</i>
</p>

<p><wxp module="wx" class="Button">
<param name="label" value="Back to the programme">
<param name="id"    value="ID_OK">
</wxp>
</p>

</center>
</body>
</html>
'''
	def __init__(self, parent):
		wx.Dialog.__init__(self, parent, -1, 'About',)
		html = wx.html.HtmlWindow(self, -1, size=(420, -1))
		if "gtk2" in wx.PlatformInfo:
			html.SetStandardFonts()
		html.SetPage(self.text)
		btn = html.FindWindowById(wx.ID_OK)
		ir = html.GetInternalRepresentation()
		html.SetSize((ir.GetWidth()+25, ir.GetHeight()+25))
		self.SetClientSize(html.GetSize())
		self.CentreOnParent(wx.BOTH)

class PE326ExportDialog(wx.Dialog):
	def __init__(self, parent, min_n, max_n, plot_func):
		wx.Dialog.__init__(self, parent, -1, 'Save as...', size=(500, -1))

		self.target_dir = os.getcwd()
		self.plot_func = plot_func

		self.min_box = wx.TextCtrl(self, -1, '%d'%(min_n))
		self.max_box = wx.TextCtrl(self, -1, '%d'%(max_n))
		count_sizer = wx.BoxSizer(wx.HORIZONTAL)
		count_sizer.Add(wx.StaticText(self, -1, 'From '), 0, wx.ALIGN_CENTRE)
		count_sizer.Add(self.min_box)
		count_sizer.Add(wx.StaticText(self, -1, ' to '), 0, wx.ALIGN_CENTRE)
		count_sizer.Add(self.max_box)
		count_box = wx.StaticBoxSizer(wx.StaticBox(self, -1, 'Output range'))
		count_box.Add(count_sizer)

		self.width_box = wx.TextCtrl(self, -1, '640', size=(50, -1))
		self.height_box = wx.TextCtrl(self, -1, '480', size=(50, -1))
		size_sizer = wx.BoxSizer(wx.HORIZONTAL)
		size_sizer.Add(wx.StaticText(self, -1, 'W '), 0, wx.ALIGN_CENTRE)
		size_sizer.Add(self.width_box)
		size_sizer.Add(wx.StaticText(self, -1, ' x H '), 0, wx.ALIGN_CENTRE)
		size_sizer.Add(self.height_box)
		size_box = wx.StaticBoxSizer(wx.StaticBox(self, -1, 'Image size'))
		size_box.Add(size_sizer)

		self.fnameformat = wx.TextCtrl(self, -1, 'plot%02d.png')
		fname_sizer = wx.BoxSizer(wx.HORIZONTAL)
		fname_sizer.Add(wx.StaticText(self, -1, 'Filename format '), 0, wx.ALIGN_CENTRE)
		fname_sizer.Add(self.fnameformat, 1, wx.EXPAND)
		fname_box = wx.StaticBoxSizer(wx.StaticBox(self, -1, 'Filename'))
		fname_box.Add(fname_sizer, 1, wx.EXPAND)

		dir_selector = wx.lib.filebrowsebutton.DirBrowseButton(
			self, -1, size=(-1, -1), changeCallback=self.dbbCallback,
			startDirectory = self.target_dir)
		dir_selector.SetValue(self.target_dir)
		dir_selector_box = wx.StaticBoxSizer(wx.StaticBox(self, -1, 'Select output directory'))
		dir_selector_box.Add(dir_selector, 1, wx.EXPAND|wx.ALL, 4)

		btn_bar = wx.StdDialogButtonSizer()
		ok_btn = wx.Button(self, wx.ID_OK)
		btn_bar.AddButton(ok_btn)
		btn_bar.AddButton(wx.Button(self, wx.ID_CANCEL))
		btn_bar.Realize()
		self.Bind(wx.EVT_BUTTON, self.Export, ok_btn)

		sz_dir = wx.BoxSizer(wx.VERTICAL)
		sz_dir.Add(count_box, 0, wx.EXPAND|wx.ALL, 4)
		sz_dir.Add(size_box, 0, wx.EXPAND|wx.ALL, 4)
		sz_dir.Add(fname_box, 0, wx.EXPAND|wx.ALL, 4)
		sz_dir.Add(dir_selector_box, 0, wx.EXPAND|wx.ALL, 4)
		sz_dir.Add(btn_bar, 0, wx.EXPAND|wx.ALL, 4)
		sz_dir.Add(wx.Size(1, 1), 1, wx.EXPAND)

		self.SetSizer(sz_dir)
		self.SetAutoLayout(True)
		self.SetBestFittingSize()

	def dbbCallback(self, event):
		self.target_dir = event.GetString()

	def ToInt(self, s, error):
		try:
			return int(s)
		except ValueError:
			dlg = wx.MessageDialog(self, error, 'Error', wx.OK | wx.ICON_ERROR)
			dlg.ShowModal()
			dlg.Destroy()
		return None

	def Export(self, evt):
		min_n = self.ToInt(self.min_box.GetLineText(0), 'Start is not valid')
		if min_n == None: return
		max_n = self.ToInt(self.max_box.GetLineText(0), 'Ending is not valid')
		if max_n == None: return
		width = self.ToInt(self.width_box.GetLineText(0), 'Width is not valid')
		if width == None: return
		height = self.ToInt(self.height_box.GetLineText(0), 'Height is not valid')
		if height == None: return
		fnameformat = self.fnameformat.GetLineText(0)

		if os.path.exists(self.target_dir):
			if not os.path.isdir(self.target_dir):
				dlg = wx.MessageDialog(self, 
					'Target is not a directory', 'Error', wx.ICON_ERROR|wx.OK)
				dlg.ShowModal()
				dlg.Destroy()
				return
		else:
			dlg = wx.MessageDialog(self, 
				'Target directory does not exist. Create?', 'Create directory', wx.ICON_QUESTION|wx.OK|wx.CANCEL)
			if dlg.ShowModal() == wx.ID_CANCEL:
				dlg.Destroy()
				return
			dlg.Destroy()
			try:
				os.mkdir(self.target_dir)
			except OSError, e:
				dlg = wx.MessageDialog(self, 
					'Failed to create directory: %s'%(str(e)), 'Create directory', wx.ICON_ERROR|wx.OK)
				dlg.ShowModal()
				dlg.Destroy()
				return

		for i in range(min_n, max_n+1):
			image = Image.new('RGB', (width, height), (255,255,255))
			self.plot_func(i, image)
			image.save(os.path.join(self.target_dir,fnameformat%(i)), 'PNG')
			
		evt.Skip()

class PE326EListBox(wx.HtmlListBox):
	def __init__(self, parent, V=0):
		wx.HtmlListBox.__init__(self, parent, -1, style=wx.BORDER_SUNKEN)
		self.V = 1.0*V
		self.es = []
		self.SetItemCount(0)

	def SetV(self, V):
		self.V = 1.0*V
		self.Refresh()

	def AppendItem(self, E):
		self.Freeze()
		self.es.append(E)
		self.SetItemCount(len(self.es))
		self.ScrollToLine(len(self.es)-1)
		self.Refresh()
		self.Thaw()

	def Refresh(self):
		if len(self.es) > 0:
			self.Freeze()
			self.RefreshLines(0, len(self.es)-1)
			self.Thaw()

	def Clear(self):
		self.SetItemCount(0)
		self.es = []
		self.Refresh()

	def OnGetItem(self, n):
		global use_eV
		if n < 0 or n >= len(self.es):
			return ''
		if use_eV:
			s = '%g eV'%(self.es[n]/eV)
		else:
			s = '%g J'%(self.es[n])
		s = 'E%d = %s'%(n+1, s)
		if self.es[n] < self.V:
			return '<font color="blue">'+s+'</font>'
		else:
			return '<font color="red">'+s+'</font>'

class PE326ComputePage(wx.NotebookPage):
	def __init__(self, parent):
		self.parent = parent
		wx.Panel.__init__(self, parent, -1)

		a_lbl = wx.StaticText(self, -1, 'a = ')
		a_txt = wx.TextCtrl(self, -1, '')
		txt_size = a_txt.GetBestSize()
		txt_size[0] = 120
		a_txt.SetMinSize(txt_size)
		a_units = wx.StaticText(self, -1, 'm')
		self.a_txt = a_txt

		w_lbl = wx.StaticText(self, -1, 'w = ')
		w_txt = wx.TextCtrl(self, -1, '')
		w_txt.SetMinSize(txt_size)
		w_units = wx.StaticText(self, -1, 'm')
		self.w_txt = w_txt

		V_lbl = wx.StaticText(self, -1, 'V_0 = ')
		V_txt = wx.TextCtrl(self, -1, '')
		V_txt.SetMinSize(txt_size)
		self.V_txt = V_txt

		V_eV = wx.RadioButton(self, -1, ' eV ')
		V_eV.SetValue(0)
		self.Bind(wx.EVT_RADIOBUTTON, self.OnVeVSelect, V_eV)
		self.V_eV = V_eV
		V_J = wx.RadioButton(self, -1, ' J ')
		V_J.SetValue(0)
		self.V_J = V_J
		self.Bind(wx.EVT_RADIOBUTTON, self.OnVJSelect, V_J)
		V_units = wx.BoxSizer(wx.HORIZONTAL)
		V_units.Add(V_eV, 0, wx.ALIGN_CENTRE|wx.LEFT|wx.RIGHT|wx.TOP, 5)
		V_units.Add(V_J, 0, wx.ALIGN_CENTRE|wx.LEFT|wx.RIGHT|wx.TOP, 5)
		global use_eV
		use_eV = True

		m_lbl = wx.StaticText(self, -1, 'm = ')
		m_txt = wx.TextCtrl(self, -1, '')
		m_txt.SetMinSize(txt_size)
		m_units = wx.StaticText(self, -1, 'kg')
		self.m_txt = m_txt

		lvls_lbl = wx.StaticText(self, -1, 'n = ')
		lvls_txt = wx.TextCtrl(self, -1, '20')
		lvls_txt.SetMinSize(txt_size)
		lvls_units = wx.StaticText(self, -1, 'energy levels')
		self.lvls_txt = lvls_txt

		self.Bind(wx.EVT_TEXT, self.UpdateValues, a_txt)
		self.Bind(wx.EVT_TEXT, self.UpdateValues, w_txt)
		self.Bind(wx.EVT_TEXT, self.UpdateValues, V_txt)
		self.Bind(wx.EVT_TEXT, self.UpdateValues, m_txt)
		self.Bind(wx.EVT_TEXT, self.UpdateValues, lvls_txt)

		space = 5
		sz_variables = wx.FlexGridSizer(cols = 3, hgap = space, vgap = space)
		sz_variables.AddMany(
			[
				a_lbl, a_txt, a_units,
				w_lbl, w_txt, w_units,
				V_lbl, V_txt, V_units,
				m_lbl, m_txt, m_units,
				lvls_lbl, lvls_txt, lvls_units,
			])

		go_button = wx.Button(self, 10, 'Start thinking')
		btn_size = go_button.GetDefaultSize()
		btn_size[0] += 30
		go_button.SetMinSize(btn_size)
		self.Bind(wx.EVT_BUTTON, self.Compute, go_button)
		self.go_button = go_button

		self.gauge = wx.Gauge(self, -1, 100)
		self.gauge.SetBezelFace(3)
		self.gauge.SetShadowWidth(3)

		sz_compute = wx.BoxSizer(wx.HORIZONTAL)
		sz_compute.Add(self.go_button, 0, wx.ALL, 3)
		sz_compute.Add(self.gauge, 1, wx.ALL, 3)
		self.sz_compute = sz_compute

		preset1_button = wx.Button(self, 20, 'Load Preset 1')
		self.Bind(wx.EVT_BUTTON, self.SelectPreset1, preset1_button)

		preset2_button = wx.Button(self, 30, 'Load Preset 2')
		self.Bind(wx.EVT_BUTTON, self.SelectPreset2, preset2_button)

		sz_presets = wx.BoxSizer(wx.HORIZONTAL)
		sz_presets.Add(preset1_button, 1, wx.EXPAND|wx.ALL, 3)
		sz_presets.Add(preset2_button, 1, wx.EXPAND|wx.ALL, 3)

		sz_topleft = wx.StaticBoxSizer(wx.StaticBox(self, -1, 'Knobs, Buttons and Stuff'), wx.VERTICAL)
		sz_topleft.Add(sz_presets, 0, wx.EXPAND)
		sz_topleft.Add(sz_variables, 0, wx.EXPAND)
		sz_topleft.Add(sz_compute, 0, wx.EXPAND)

		self.minigraph = MiniGraph(self)

		self.E_list = PE326EListBox(self)
		self.func_cache = []
		self.data_cache = []

		self.LogText = wx.TextCtrl(self, -1, '',
			style=wx.TE_MULTILINE|wx.TE_RICH2|wx.TE_READONLY|wx.VSCROLL|wx.HSCROLL)

		sz_top = wx.BoxSizer(wx.HORIZONTAL)
		sz_top.Add(sz_topleft, 0, wx.ALL, 2)
		sz_top.Add(self.minigraph, 1, wx.EXPAND|wx.ALL, 2)
		sz_top.Add(self.E_list, 1, wx.EXPAND|wx.ALL, 2)

		sz_bottom = wx.StaticBoxSizer(wx.StaticBox(self, -1, 'Good Things and Bad Things'), wx.VERTICAL)
		sz_bottom.Add(self.LogText, 1, wx.EXPAND)

		clear_log_btn = wx.Button(self, 40, 'Erase working out')
		self.Bind(wx.EVT_BUTTON, self.ClearLog, clear_log_btn)
		about_btn = wx.Button(self, 50, 'Who\'s who?')
		self.Bind(wx.EVT_BUTTON, self.About, about_btn)
		self.errors_chk = wx.CheckBox(self, -1, 'Bloviate')
		self.errors_chk.SetValue(0)
		exit_btn = wx.Button(self, wx.ID_EXIT)
		self.Bind(wx.EVT_BUTTON, self.CloseMe, exit_btn)
		#self.plot_imaginary_chk = wx.CheckBox(self, -1, 'Imagine things',
			#style=wx.CHK_3STATE|wx.CHK_ALLOW_3RD_STATE_FOR_USER)
		#self.plot_imaginary_chk.Set3StateValue(PE326Computer.plot_imaginary)
		#self.Bind(wx.EVT_CHECKBOX, self.UpdateImaginary, self.plot_imaginary_chk)

		self.export_btn = wx.Button(self, wx.ID_SAVE)
		self.export_btn.Disable()
		self.Bind(wx.EVT_BUTTON, self.Export, self.export_btn)

		sz_bottombar = wx.StaticBoxSizer(wx.StaticBox(self, -1, 'Dooverlackies and Whatsits'), wx.HORIZONTAL)
		sz_bottombar.Add(clear_log_btn, 0, wx.ALL, 4)
		sz_bottombar.Add(about_btn, 0, wx.ALL, 4)
		sz_bottombar.Add(self.errors_chk, 0, wx.TOP|wx.BOTTOM, 10)
		#sz_bottombar.Add(self.plot_imaginary_chk, 0, wx.ALL, 4)
		sz_bottombar.Add(wx.Size(1,1), 1, wx.EXPAND)
		sz_bottombar.Add(self.export_btn, 0, wx.ALL, 4)
		sz_bottombar.Add(exit_btn, 0, wx.ALL, 4)

		sz_outer = wx.BoxSizer(wx.VERTICAL)
		sz_outer.Add(sz_top, 0, wx.EXPAND|wx.ALL, 2)
		sz_outer.Add(sz_bottom, 1, wx.EXPAND|wx.ALL, 2)
		sz_outer.Add(sz_bottombar, 0, wx.EXPAND|wx.ALL, 2)

		self.SetSizer(sz_outer)
		self.SetAutoLayout(True)

		self.SelectPreset1()
		self.UpdateMinigraph()

		self.running = False
		self.HideThinker()

		self.last_parameters = None

		self.last_E_num = 0

	def HideThinker(self):
		self.gauge.Disable()
		self.gauge.SetValue(0)

	def ShowThinker(self):
		self.gauge.Enable()

	def ValuesAreSane(self):
		try:               a = float(self.a_txt.GetLineText(0))
		except ValueError: return 'a is not a real number.'
		try:               w = float(self.w_txt.GetLineText(0))
		except ValueError: return 'w is not a real number.'
		try:               V = float(self.V_txt.GetLineText(0))
		except ValueError: return 'V_0 is not a real number.'
		try:               m = float(self.m_txt.GetLineText(0))
		except ValueError: return 'm is not a real number.'
		try:               n = int(self.lvls_txt.GetLineText(0))
		except ValueError: return 'Number of energy levels is not an integer.'
		global use_eV
		if use_eV:
			V = eVToJ(V)
		if V > 1e6 * eV: return 'Insane value of V.'
		if a <  0: return 'a cannot be negative.'
		if a == 0: return 'There is no well.'
		if w <  0: return 'w cannot be negative.'
		if w >  a: return 'w must be less than a.'
		if n <  1: return 'Number of energy levels must be positive.'
		if m == 0: return 'm cannot be 0.'
		if m <  0: return 'm cannot be negative.'
		if V <  0: return 'V cannot be negative.'
		return None

	def CloseMe(self, event):
		self.parent.GetParent().Close()

	def ClearLog(self, event):
		self.LogText.Clear()

	def Export(self, event):
		export_dlg = PE326ExportDialog(self, 1, self.last_E_num, self.PlotFunc)
		export_dlg.ShowModal()
		export_dlg.Destroy()

	def About(self, event):
		about_box = AboutBox(self)
		about_box.ShowModal()
		about_box.Destroy()

	def UpdateImaginary(self, event):
		PE326ComputePage.plot_imaginary = self.plot_imaginary_chk.Get3StateValue()

	def UpdateValues(self, event):
		self.UpdateMinigraph()

	def AddE(self, E_num, E_val, func, func_data):
		self.E_list.AppendItem(E_val)
		if E_num == 1:
			self.export_btn.Enable()
		self.last_E_num = E_num
		self.func_cache.append(func)
		self.data_cache.append(func_data)

	def RefreshE(self):
		self.E_list.Refresh()

	def UpdateMinigraph(self):
		s = self.ValuesAreSane()
		if s == None:
			a = float(self.a_txt.GetLineText(0))
			w = float(self.w_txt.GetLineText(0))
			V = float(self.V_txt.GetLineText(0))
			global use_eV
			if use_eV:
				V = eVToJ(V)
			self.minigraph.UpdateValues(a, w, V)
			self.minigraph.SetSanity(True)
		else:
			self.minigraph.SetSanity(False, s)
		self.minigraph.Refresh()

	def Log(self, str, type=0):
		if self.errors_chk.GetValue() == 0 and type == 0:
			return
		pos = self.LogText.GetInsertionPoint()
		self.LogText.AppendText(str)
		if type == -1: # ERROR
			self.LogText.SetStyle(pos, self.LogText.GetInsertionPoint(), wx.TextAttr('RED'))
		if type == 1: # WARNING
			self.LogText.SetStyle(pos, self.LogText.GetInsertionPoint(), wx.TextAttr('ORANGE'))
		if type == 2: # GOOD
			self.LogText.SetStyle(pos, self.LogText.GetInsertionPoint(), wx.TextAttr('GREEN'))
		if type == 3: # BLUE
			self.LogText.SetStyle(pos, self.LogText.GetInsertionPoint(), wx.TextAttr('BLUE'))
	
	def OnVeVSelect(self, event):
		global use_eV
		if use_eV:
			return
		use_eV = True
		old_val = self.V_txt.GetLineText(0)
		new_val = JToeV(old_val)
		if new_val != None and new_val < 1e10:
			self.V_txt.Clear()
			self.V_txt.write('%.15g'%(new_val))
		self.UpdateValues(event)
		self.RefreshE()

	def OnVJSelect(self, event):
		global use_eV
		if not use_eV:
			return
		use_eV = False
		old_val = self.V_txt.GetLineText(0)
		new_val = eVToJ(old_val)
		if new_val != None and new_val > 1e-30:
			self.V_txt.Clear()
			self.V_txt.write('%.15g'%(new_val))
		self.UpdateValues(event)
		self.RefreshE()

	def SelectPreset1(self, event = None):
		self.V_eV.SetValue(1)
		global use_eV
		use_eV = True
		self.a_txt.Clear()
		self.a_txt.write('10e-9')
		self.w_txt.Clear()
		self.w_txt.write('4.0e-9')
		self.V_txt.Clear()
		self.V_txt.write('0.5')
		self.m_txt.Clear()
		self.m_txt.write('9.11e-31')
		self.lvls_txt.Clear()
		self.lvls_txt.write('20')

	def SelectPreset2(self, event = None):
		self.V_eV.SetValue(1)
		global use_eV
		use_eV = True
		self.a_txt.Clear()
		self.a_txt.write('10e-9')
		self.w_txt.Clear()
		self.w_txt.write('0.1e-9')
		self.V_txt.Clear()
		self.V_txt.write('0.5')
		self.m_txt.Clear()
		self.m_txt.write('9.11e-31')
		self.lvls_txt.Clear()
		self.lvls_txt.write('30')

	def Abort(self):
		if self.running:
			self.parent.computer.Stop()
			self.Log("\nAborting...\n")

	def Compute(self, event):
		# are we already running? if so, stop
		if self.running:
			self.Abort()
			return

		str = self.ValuesAreSane()
		if str != None:
			self.Log(str+'\n', -1)
			return

		a = float(self.a_txt.GetLineText(0))
		w = float(self.w_txt.GetLineText(0))
		V = float(self.V_txt.GetLineText(0))
		m = float(self.m_txt.GetLineText(0))
		n = int(self.lvls_txt.GetLineText(0))

		global use_eV
		if use_eV:
			V = eVToJ(V)

		self.go_button.SetLabel('Stop thinking')

		self.gauge.SetRange(n)
		self.gauge.SetValue(0)
		self.ShowThinker()

		self.running = True

		cont = False
		clear_pages = True
		if self.last_parameters != None and \
				self.last_parameters['a'] == a and \
				self.last_parameters['w'] == w and \
				self.last_parameters['V'] == V and \
				self.last_parameters['m'] == m and \
				self.last_parameters['n'] > 0:
			cont = True
			clear_pages = False
		else:
			self.last_parameters = {}
			self.last_parameters['a'] = a
			self.last_parameters['w'] = w
			self.last_parameters['V'] = V
			self.last_parameters['m'] = m
			self.last_parameters['n'] = 0
			self.E_list.SetV(V)

		if clear_pages:
			self.E_list.Clear()
			self.func_cache = []
			self.data_cache = []
			self.export_btn.Disable()

		self.parent.compute(a, w, V, m, n, cont, clear_pages)

	def SetGauge(self, n):
		self.gauge.SetValue(n)

	def ComputeDone(self):
		self.running = False
		self.go_button.SetLabel('Start thinking')
		self.HideThinker()

	def PlotFunc(self, E_num, image):
		def ptToX(x):
			viewport_w = width - margin_left - margin_right
			return int(margin_left + ((x - start) * viewport_w) / (end - start))

		def ptToY(y):
			viewport_h = height - margin_top - margin_bottom
			return int(margin_top + viewport_h - (((y - ymin) * viewport_h) / (ymax - ymin)))

		func  = self.func_cache[E_num-1]
		fdata = self.data_cache[E_num-1]
		[a, w, V, m, A, B, C, D, E, even] = fdata
		start = -a
		end   =  a

		margin_left    = 15
		margin_right   = 15
		margin_top     = 50
		margin_bottom  = 25

		width, height = image.size

		dc = ImageDraw.Draw(image)

		try:
			small_font = ImageFont.truetype('arial.ttf', height/40)
			big_font = ImageFont.truetype('arial.ttf', height/30)
		except IOError:
			small_font = big_font = ImageFont.load_default()

		# Energy level
		global use_eV
		str = 'E%d = '%(E_num)
		if use_eV:
			str += '%g eV'%(JToeV(E))
		else:
			str += '%g J'%(E)
		dc.text((2*margin_left, 2), str, fill=(0,0,0), font=big_font)
		tw, th = dc.textsize(str, font=big_font)

		if use_eV:
			vstr = '%g eV'%(JToeV(V))
		else:
			vstr = '%g J'%(V)
		bottom_str = 'Parameters: a = %g   w = %g   V = %s   m = %g'%(a, w, V, m)
		dc.text((2*margin_left, 4 + th), bottom_str, fill=(0,0,0), font=big_font)
		tw2, th2 = dc.textsize(bottom_str, font=big_font)
		margin_top = 8 + th + th2

		# Calculate graph points
		x_range = end - start
		incr = 1.0*x_range/(width - margin_left - margin_right)
		ymin = 0
		ymax = 0
		x_pt_values = []
		fpoints = []
		x = start
		while x < end:
			y = func(x, fdata)
			fpoints.append((x, y))
			if y < ymin: ymin = y
			if y > ymax: ymax = y
			x += incr

		if ymin == 0 and ymax == 0:
			ymin = -1
			ymax = 1

		# Convert to points on canvas
		draw_points = map(lambda (x, y): (ptToX(x), ptToY(y)), fpoints)

		if w != None:
			# Barrier lines
			w_left  = (ptToX(-w), ptToY(ymin), ptToX(-w), ptToY(ymax))
			w_right = (ptToX( w), ptToY(ymin), ptToX( w), ptToY(ymax))

		# Box
		dc.rectangle([margin_left-1, margin_top-1, 
				width  - margin_right  - 1,
				height - margin_bottom - 1], outline=(0,0,0))

		if w != 0:
			# Barrier
			dc.line([w_left[0], w_left[1], w_left[2], w_left[3]], fill=(128,128,128))
			dc.line([w_right[0], w_right[1], w_right[2], w_right[3]], fill=(128,128,128))

		# Axes
		dc.line([ptToX(0), margin_top, ptToX(0), height-margin_bottom], fill=(0,0,0))
		dc.line([margin_left, ptToY(0), width-margin_right, ptToY(0)], fill=(0,0,0))

		# X axis ticks
		xtickstep = 10.0**int(math.log10(end - start)-1)
		while (end - start)/xtickstep > 20: xtickstep *= 10
		while (end - start)/xtickstep < 3: xtickstep /= 10
		x = int(start/xtickstep)*xtickstep
		last_w = None
		while x < end:
			dc.line([ptToX(x), ptToY(0)-5, ptToX(x), ptToY(0)+5], fill=(0,0,0))
			if last_w == None or ptToX(x) - last_w > 90:
				if last_w != None and abs(x) >= xtickstep:
					dc.text((ptToX(x)-8, ptToY(0)+7), '%g'%(x), fill=(0,0,0), font=small_font)
				last_w = ptToX(x)
			x += xtickstep

		# Y axis ticks
		ytickstep = 10.0**int(math.log10(ymax - ymin)-1)
		while (ymax - ymin)/ytickstep > 20: ytickstep *= 10
		while (ymax - ymin)/ytickstep < 3: ytickstep /= 10
		y = int(ymin/ytickstep)*ytickstep
		last_h = None
		while y < ymax:
			dc.line([ptToX(0)-5, ptToY(y), ptToX(0)+5, ptToY(y)], fill=(0,0,0))
			if last_h == None or last_h - ptToY(y) > 80:
				if last_h != None and abs(y) >= ytickstep:
					dc.text([ptToX(0)+7, ptToY(y)-5], '%g'%(y), fill=(0,0,0), font=small_font)
				last_h = ptToY(y)
			y += ytickstep

		# Plot points
		if E == None or E > V:
			colour = (255,0,0)
		else:
			colour = (0,0,255)
		dc.line(draw_points, fill=colour)

		del dc

class PE326Graph(wx.Panel):
	def __init__(self, parent, n, f, fdata):
		wx.Panel.__init__(self, parent, -1, style=wx.FULL_REPAINT_ON_RESIZE|wx.TAB_TRAVERSAL|wx.NO_BORDER)
		self.SetBackgroundColour(wx.WHITE)
		self.Bind(wx.EVT_PAINT, self.OnPaint)
		self.n = n
		self.f = f
		self.fdata = fdata
		self.draw_points = None
		self.Bind(wx.EVT_SIZE, self.Resize)
		[self.a, self.w, self.V, self.m, self.A, self.B, self.C, self.D, self.E, self.even] = fdata
		self.start = -self.a
		self.end   =  self.a

		self.margin_left    = 15
		self.margin_right   = 15
		self.margin_top     = 25
		self.margin_bottom  = 25

	def Resize(self, evt):
		self.SetSize(self.GetMaxSize())
		self.draw_points = None

	def ptToX(self, x):
		viewport_w = self.width - self.margin_left - self.margin_right
		return int(self.margin_left + ((x - self.start) * viewport_w) / (self.end - self.start))

	def ptToY(self, y):
		viewport_h = self.height - self.margin_top - self.margin_bottom
		return int(self.margin_top + viewport_h - (((y - self.ymin) * viewport_h) / (self.ymax - self.ymin)))

	def Precalculate(self):
		(self.width, self.height) = self.GetSizeTuple()
		x_range = self.end - self.start
		incr = 1.0*x_range/(self.width - self.margin_left - self.margin_right)
		self.ymin = 0
		self.ymax = 0
		x_pt_values = []
		self.fpoints = []
		x = self.start
		while x < self.end:
			y = self.f(x, self.fdata)
			self.fpoints.append((x, y))
			if y < self.ymin: self.ymin = y
			if y > self.ymax: self.ymax = y
			x += incr

		if self.ymin == 0 and self.ymax == 0:
			self.ymin = -1
			self.ymax = 1

		# Convert to points on canvas
		self.draw_points = map(lambda (x, y): (self.ptToX(x), self.ptToY(y)), self.fpoints)

		if self.w != None:
			# Barrier lines
			self.w_left  = (self.ptToX(-self.w), self.ptToY(self.ymin), self.ptToX(-self.w), self.ptToY(self.ymax))
			self.w_right = (self.ptToX( self.w), self.ptToY(self.ymin), self.ptToX( self.w), self.ptToY(self.ymax))

	def Plot(self, dc):
		if self.draw_points == None:
			self.Precalculate()

		dc.BeginDrawing()

		# Energy level
		global use_eV
		if self.n != None:
			str = 'E%d = '%(self.n)
			if use_eV:
				str += '%g eV'%(JToeV(self.E))
			else:
				str += '%g J'%(self.E)
			dc.DrawText(str, 2*self.margin_left, 2)

		# Box
		dc.DrawRectangle(self.margin_left-1, self.margin_top-1, 
				self.width  - self.margin_left - self.margin_right  + 2,
				self.height - self.margin_top  - self.margin_bottom + 2)

		if self.w != None:
			# Barrier
			dc.SetPen(wx.Pen('BLACK', 1, wx.DOT))
			dc.DrawLine(self.w_left[0], self.w_left[1], self.w_left[2], self.w_left[3])
			dc.DrawLine(self.w_right[0], self.w_right[1], self.w_right[2], self.w_right[3])

		# Axes
		dc.SetPen(wx.Pen('BLACK', 1, wx.SOLID))
		dc.DrawLine(self.ptToX(0), self.margin_top, self.ptToX(0), self.height-self.margin_top)
		dc.DrawLine(self.margin_left, self.ptToY(0), self.width-self.margin_left, self.ptToY(0))

		# X axis ticks
		xtickstep = 10.0**int(math.log10(self.end - self.start)-1)
		while (self.end - self.start)/xtickstep > 20: xtickstep *= 10
		while (self.end - self.start)/xtickstep < 3: xtickstep /= 10
		x = int(self.start/xtickstep)*xtickstep
		last_w = None
		while x < self.end:
			dc.DrawLine(self.ptToX(x), self.ptToY(0)-5, self.ptToX(x), self.ptToY(0)+5)
			if last_w == None or self.ptToX(x) - last_w > 90:
				if last_w != None and abs(x) >= xtickstep:
					dc.DrawText('%g'%(x), self.ptToX(x)-8, self.ptToY(0)+7)
				last_w = self.ptToX(x)
			x += xtickstep

		# Y axis ticks
		ytickstep = 10.0**int(math.log10(self.ymax - self.ymin)-1)
		while (self.ymax - self.ymin)/ytickstep > 20: ytickstep *= 10
		while (self.ymax - self.ymin)/ytickstep < 3: ytickstep /= 10
		y = int(self.ymin/ytickstep)*ytickstep
		last_h = None
		while y < self.ymax:
			dc.DrawLine(self.ptToX(0)-5, self.ptToY(y), self.ptToX(0)+5, self.ptToY(y))
			if last_h == None or last_h - self.ptToY(y) > 80:
				if last_h != None and abs(y) >= ytickstep:
					dc.DrawText('%g'%(y), self.ptToX(0)+7, self.ptToY(y)-5)
				last_h = self.ptToY(y)
			y += ytickstep

		# Plot points
		if self.E == None or self.E > self.V:
			dc.SetPen(wx.Pen('RED', 1))
		else:
			dc.SetPen(wx.Pen('BLUE', 1))
		dc.DrawLines(self.draw_points)

		dc.EndDrawing()

	def OnPaint(self, evt):
		dc = wx.PaintDC(self)
		dc.Clear()
		self.Plot(dc)

class PE326Window(wx.Notebook):
	def __init__(self, parent):
		wx.Notebook.__init__(self, parent, -1)
		self.computePage = PE326ComputePage(self)
		self.AddPage(self.computePage, "Start Here")
		self.Bind(PE326Computer.EVT_UPDATE_E, self.OnUpdate)
		self.Bind(PE326Computer.EVT_COMPLETE, self.Completed)
		self.Bind(PE326Computer.EVT_LOGMESSAGE, self.LogMessage)

	def clear_pages(self):
		n = self.GetPageCount()
		for i in range(1, n):
			self.DeletePage(1)

	def compute(self, a, w, v, m, n, cont, clear_pages=True):
		if clear_pages:
			self.clear_pages()
		self.computer = PE326Computer.PE326Computer(self, a, w, v, m, n, cont)

	def OnUpdate(self, evt):
		self.AddPage(PE326Graph(self, evt.E_num, evt.func, evt.func_data), 
			'E%d'%(evt.E_num))
			#'E%d = %g'%(evt.E_num, evt.E))
		xa, xw, xv, xm, xA, xB, xC, xD, xE, xeven = evt.func_data
		self.computePage.last_parameters['n'] = evt.E_num
		self.computePage.last_parameters['E'] = xE
		self.computePage.SetGauge(evt.E_num)
		self.computePage.AddE(evt.E_num, xE, evt.func, evt.func_data)

	def Completed(self, evt):
		self.computePage.ComputeDone()
		try:
			del self.computer
		except AttributeError:
			pass

	def LogMessage(self, evt):
		try:
			type = evt.type
		except AttributeError:
			self.computePage.Log(evt.str)
			return
		self.computePage.Log(evt.str, evt.type)

class PE326App(wx.App):
	def __init__(self):
		wx.App.__init__(self, redirect=False)

	def OnInit(self):
		frame = wx.Frame(None, -1, "Quantum Computer 3.26 (Personal Edition)", pos=(50,50), size=(500,400),
			style=wx.DEFAULT_FRAME_STYLE)
		frame.Bind(wx.EVT_CLOSE, self.OnCloseFrame)

		win = PE326Window(frame)
		frame.SetSize((640,480))
		win.SetFocus()
		self.window = win
		self.SetTopWindow(frame)
		self.frame = frame
		frame.Show(True)
		return True

	def OnExitApp(self, evt):
		self.frame.Close(True)

	def OnCloseFrame(self, evt):
		self.window.computePage.Abort()
		evt.Skip()

PE326App().MainLoop()

