#
# PE326 Assignment
# 
# Infinite Quantum Well of width 2*a with a potential step of width 2*w in the
# center.
# 
# This file contains the main computation engine. It runs as a thread (but uses
# global variables so make sure only ONE ever runs at a time), called from the
# GUI.
# 

from math import *
import copy
import sys
import thread
import wx, wx.lib.newevent
import traceback

j = 1j
h = 6.6260755e-34
hbar = h/(2 * pi)
eV = 1.60211773e-19

E_LT_V = 0
E_GT_V = 1
EVEN = 0
ODD = 1

matrices = \
 [
  [
   [ # E < V, EVEN
	[ lambda:        -1 * e**(1 * k2() * w), lambda:        -1 * e**(-1 * k2() * w), lambda:            e**(+j * k1() * w), lambda:             e**(-j * k1() * w) ],
	[ lambda:                            -1, lambda:                              1, lambda:                             0, lambda:                              0 ],
	[ lambda: -1 * k2() * e**(1 * k2() * w), lambda:  1 * k2() * e**(-1 * k2() * w), lambda: j * k1() * e**(+j * k1() * w), lambda: -j * k1() * e**(-j * k1() * w) ],
	[ lambda:                             0, lambda:                              0, lambda:            e**(+j * k1() * a), lambda:             e**(-j * k1() * a) ],
   ],
   [ # E < V, ODD
	[ lambda:        -1 * e**(1 * k2() * w), lambda:        -1 * e**(-1 * k2() * w), lambda:            e**(+j * k1() * w), lambda:             e**(-j * k1() * w) ],
	[ lambda:                             1, lambda:                              1, lambda:                             0, lambda:                              0 ],
	[ lambda: -1 * k2() * e**(1 * k2() * w), lambda:  1 * k2() * e**(-1 * k2() * w), lambda: j * k1() * e**(+j * k1() * w), lambda: -j * k1() * e**(-j * k1() * w) ],
	[ lambda:                             0, lambda:                              0, lambda:            e**(+j * k1() * a), lambda:             e**(-j * k1() * a) ],
   ],
  ],
  [
   [ # E > V, EVEN
	[ lambda:        -1 * e**(j * k2() * w), lambda:        -1 * e**(-j * k2() * w), lambda:            e**(+j * k1() * w), lambda:             e**(-j * k1() * w) ],
	[ lambda:                            -j, lambda:                              j, lambda:                             0, lambda:                              0 ],
	[ lambda:                             0, lambda:                              0, lambda:            e**(+j * k1() * a), lambda:             e**(-j * k1() * a) ],
	[ lambda: -j * k2() * e**(j * k2() * w), lambda:  j * k2() * e**(-j * k2() * w), lambda: j * k1() * e**(+j * k1() * w), lambda: -j * k1() * e**(-j * k1() * w) ],
   ],
   [ # E > V, ODD
	[ lambda:        -1 * e**(j * k2() * w), lambda:        -1 * e**(-j * k2() * w), lambda:            e**(+j * k1() * w), lambda:             e**(-j * k1() * w) ],
	[ lambda:                             1, lambda:                              1, lambda:                             0, lambda:                              0 ],
	[ lambda:                             0, lambda:                              0, lambda:            e**(+j * k1() * a), lambda:             e**(-j * k1() * a) ],
	[ lambda: -j * k2() * e**(j * k2() * w), lambda:  j * k2() * e**(-j * k2() * w), lambda: j * k1() * e**(+j * k1() * w), lambda: -j * k1() * e**(-j * k1() * w) ],
   ]
  ]
 ]

k1 = lambda: sqrt(2 * m * E) / hbar
k2 = lambda: sqrt(2 * m * abs(V-E)) / hbar

global LogFunc
global E, a, w, m, V, A, B, C, D, epsilon
epsilon = 1e-5

plot_imaginary = 2 # plot magnitude of the wave function, with the sign chosen appropriately

def det(matrix):
	'''Returns the determinant of a square matrix with functors for entries.'''
	size = len(matrix)
	# Ensure it's square
	if len(matrix[0]) != size:
		cols = len(matrix[0])
		raise Exception("Cannot get determinant of %dx%d matrix"%(size,cols))
	# Hard-code the 2x2 case.
	if size == 2:
		return matrix[0][0]()*matrix[1][1]() - matrix[0][1]() * matrix[1][0]()
	# Else do the thing.
	ans = 0
	for col in range(0,size):
		M = []
		for i in matrix[1:]:
			M.append(i[:col]+i[col+1:])
		ans += matrix[0][col]() * (-1)**(col) * det(M)
	return ans

def det_setE(matrix, newE):
	'''Miscellaneous function that sets E to the value given, then returns the
	determinant of the matrix.'''
	global E
	E = newE
	return det(matrix)


def is_sensible_range(f, lo, hi, sample_points = 50):
	'''Determines if a function goes up once and down once within the range
	[lo, hi] at a certain number of sample points.
	Returns a 2-tuple of booleans.
	If the first tuple is true, the condition holds.
	Otherwise, if the second tuple is True, the function was monotonic in the
	range [lo, hi] (ie, range too small).
	If the second tuple is False, the function went up and down at least once
	in the range[lo, hi] (ie, range too large).'''
	x = lo
	going_up = -2
	changed = False
	step = (hi - lo) / sample_points
	#print '\t\tis_sensible_range(f, %s, %s)'%(str(lo),str(hi))
	while x < hi:
		#print "Trying with", x, 'giving', f(x)
		res = f(x)
		if going_up == -2:
			# Don't have any data points yet. Will do next time.
			going_up = -1
		elif going_up == -1:
			# Now have a data point. Figure out which way we're going
			if prev_res < res:
				going_up = 1
			else:
				going_up = 0
		elif going_up == 0:
			if prev_res < res:
				if changed:
					return (False, False) # 2 changes of direction already. Quit it.
				else:
					changed = True
					going_up = 1
		elif going_up == 1:
			if prev_res > res:
				if changed:
					return (False, False) # 2 changes of direction already. Quit it.
				else:
					changed = True
					going_up = 0
		prev_res = res
		x += step
	# All went well.
	if changed:
		return (True, True)
	return (False, True)

def find_sensible_increment(f, start, start_incr):
	'''Determines an increment suitable to start solving for E with.'''
	lo = start
	hi = start + start_incr
	#print '\t\tfind_sensible_increment(%s,%s)'%(str(lo),str(hi))
	sample_points = 50.
	while True:
		(good, too_small) = is_sensible_range(f, lo, hi, sample_points)
		if good:
			return (hi-lo)/sample_points/10
		if too_small:
			#print '\ttoo small', lo, hi
			hi *= 2.
		else:
			#print '\ttoo big', lo, hi
			hi /= 11.

def converged(a, b):
	'''Determines if a and b agree to 20 significant figures.'''
	return abs(log(a) - log(b)) < 1e-20

def solve_for_zero(f, lo, hi):
	# FIXME ... is this really doing the right thing?
	spincount = 0
	while True:
		#print "Readjusting to %g and %g"%(lo,hi)
		incr = (hi-lo)/100.
		x = lo
		prev_ans = None
		ans = f(x)
		#print '%.80e on\n\t%.80e'%(incr, x)
		old_x = x
		while prev_ans == None or (ans > 0 and prev_ans > 0) or (ans < 0 and prev_ans < 0):
			prev_ans = ans
			x += incr
			if x == old_x: break
			ans = f(x)
			old_x = x
		lo = x - incr
		hi = x + incr
		if converged(lo, hi): break
		spincount += 1
		if spincount > 1000:
			LogFunc('WARNING: spun out on solve_for_zero %.15g to %.15g\n'%(lo,hi), 1)
			break
	#print 'And out with %.15g'%x
	return x

def xsolve_for_zero(f, lo, hi):
	'''Solve the equation f equal to zero between lo and hi using a modified Newton-Raphson.'''
	x = oldx = lo
	spincount = 0
	fhi = f(hi)
	flo = f(lo)
	dydx = (fhi-flo)/(hi-lo)
	while True:
		y = f(x)
		x = x - y/dydx
		if converged(x, oldx): break
		oldx = x
		spincount += 1
		if spincount > 100:
			LogFunc('WARNING: spun out converging between %.50e and %.50e\n'%(lo, hi), 1)
			break
	return x

def solve(f, start_x, start_increment, limit):
	'''Solve for the function f(x) = 0 starting from start_x.'''
	# Firstly find a sensible range to search in.
	x = start_x
	incr = find_sensible_increment(f, x, start_increment)
	LogFunc(' .')
	sys.stdout.flush()

	# Find the point where the function crosses the axis.
	prev_y = None
	while True:
		y = f(x)
		if prev_y != None and (y > 0 and prev_y < 0) or (y < 0 and prev_y > 0):
			break
		prev_y = y
		x += incr
		if limit != None and x > limit:
			LogFunc('!')
			sys.stdout.flush()
			return (None, start_increment, limit)
	LogFunc('.')
	sys.stdout.flush()

	# Solution lies between x-incr. and x. Go searching.
	res = solve_for_zero(f, x-incr, x)
	LogFunc('.')
	return (res, x, incr)

def prettyprint_matrix(matrix):
	for row in matrix:
		for col in row:
			print "%s\t"%(col),
		print

def add_row(dst, src, scale):
	'''Adds two rows of a matrix together. Specfically dst = dst + src * scale'''
	if len(dst) != len(src):
		raise Exception("Can't add rows of length %d and %d"%(len(dst),len(src)))
	for i in range(0,len(dst)):
		dst[i] += src[i] * scale

def mult_row(row, scale):
	'''Multiply a row by a given scaling factor.'''
	for i in range(0,len(row)):
		row[i] *= scale

def make_echelon(matrix):
	'''Convert a matrix into echelon form.'''
	for row_num in range(0,len(matrix)):
		row = matrix[row_num]
		for i in range(0,row_num):
			if matrix[i][i] != 0:
				add_row(row, matrix[i], -row[i]/matrix[i][i])
				row[i] = 0.0+0j
		# HACK: our last row rarely comes out to exactly zero, so make it so.
		if row_num == len(matrix)-1:
			if w != 0 and abs(matrix[row_num][row_num]) > 1:
				LogFunc('WARNING: substantial leftovers when determining nullspace:\n', -1)
				for r in matrix:
					LogFunc('\t', 1)
					for c in r:
						LogFunc('%s\t'%(str(c)), 1)
					LogFunc('\n', 1)
			matrix[row_num] = [0]*4
			break
		if row[row_num] != 0:
			mult_row(row, 1./row[row_num])
		row[row_num] = 1+0j
		row_num += 1
	# Now really make it echelon form
	for i in range(len(matrix)-1,-1,-1):
		if matrix[i][i] == 0:
			continue
		for k in range(0,i):
			add_row(matrix[k], matrix[i], -matrix[k][i])
			matrix[k][i] = 0

def find_nullspace(matrix):
	#prettyprint_matrix(matrix)
	make_echelon(matrix)
	ns = [1, -matrix[2][3], -matrix[1][3], -matrix[0][3]]
	#print ns
	return ns

def integrate(f, fdata, start, end, steps):
	integral = 0.0
	x = start
	incr = (end-start)/steps
	while x < end:
		y = f(x+(incr/2.0), fdata)
		integral += incr * y
		x += incr
	return integral

def sgn(x):
	if x == 0: return 0
	if x < 0: return -1
	return 1

def choose_complex_portion(x, fdata):
	if plot_imaginary == 0:
		return x.real
	elif plot_imaginary == 1:
		return x.imag
	elif plot_imaginary == 2:
		(xa, xw, xV, xm, xA, xB, xC, xD, xE, xeven) = fdata
		if xE < xV or xeven:
			return x.real
		return x.imag

(UpdateEEvent, EVT_UPDATE_E) = wx.lib.newevent.NewEvent()
(CompleteEvent, EVT_COMPLETE) = wx.lib.newevent.NewEvent()
(LogMessage, EVT_LOGMESSAGE) = wx.lib.newevent.NewEvent()

class PE326Computer:
	def __init__(self, parent, a, w, v, m, n, cont=False):
		global LogFunc
		LogFunc = lambda s, t=0: self.Log(s, t)
		self.parent = parent
		self.a = a
		self.w = w
		self.v = v
		self.m = m
		self.n = n
		self.cont = cont
		self.stop_now = False
		thread.start_new_thread(self.Run, ())

	def NotifyOfE(self, E_num_val, E_val, f, fdata):
		evt = UpdateEEvent(E_num = E_num_val, E = E_val, func = f, func_data = fdata)
		wx.PostEvent(self.parent, evt)

	def NotifyOfComplete(self):
		evt = CompleteEvent()
		wx.PostEvent(self.parent, evt)

	def Log(self, s, t=None):
		if t == None:
			evt = LogMessage(str = s)
		else:
			evt = LogMessage(str = s, type = t)
		wx.PostEvent(self.parent, evt)

	def func_e_lt_v(self, x, fdata):
		(xa, xw, xV, xm, xA, xB, xC, xD, xE, xeven) = fdata
		xk1 = sqrt(2 * xm * xE) / hbar
		xk2 = sqrt(2 * xm * abs(xV - xE)) / hbar
		ans = 0+0j
		if x < -xa:
			ans = 0+0j
		elif x < -xw:
			ans = complex(xA * e**(1j * xk1 * x) + xB * e**(-1j * xk1 * x))
		elif x < 0:
			ans = complex(xC * e**(1  * xk2 * x) + xD * e**(-1  * xk2 * x))
		elif x < xw:
			if xeven:
				ans = complex(xC * e**(1  * xk2 * -x) + xD * e**(-1  * xk2 * -x))
			else:
				ans = complex(-(xC * e**(1  * xk2 * -x) + xD * e**(-1  * xk2 * -x)))
		elif x < xa:
			if xeven:
				ans = complex(xA * e**(1j * xk1 * -x) + xB * e**(-1j * xk1 * -x))
			else:
				ans = complex(-(xA * e**(1j * xk1 * -x) + xB * e**(-1j * xk1 * -x)))
		#if (abs(ans.real) != 0 and abs(ans.imag/ans.real) > 10):
		#	LogFunc("Hmmm, I think I'm imagining something: %g vs %g\n"%(ans.real, ans.imag), -1)
		return choose_complex_portion(ans, fdata)

	def func_e_gt_v(self, x, fdata):
		(xa, xw, xV, xm, xA, xB, xC, xD, xE, xeven) = fdata
		xk1 = sqrt(2 * xm * xE) / hbar
		xk2 = sqrt(2 * xm * abs(xV - xE)) / hbar
		ans = 0+0j
		if x < -xa:
			ans = 0+0j
		elif x < -xw:
			ans = complex(xA * e**(1j * xk1 * x) + xB * e**(-1j * xk1 * x))
		elif x < 0:
			ans = complex(xC * e**(1j * xk2 * x) + xD * e**(-1j * xk2 * x))
		elif x < xw:
			if xeven:
				ans = complex(xC * e**(1j * xk2 * -x) + xD * e**(-1j * xk2 * -x))
			else:
				ans = complex(-(xC * e**(1j * xk2 * -x) + xD * e**(-1j * xk2 * -x)))
		elif x < xa:
			if xeven:
				ans = complex(xA * e**(1j * xk1 * -x) + xB * e**(-1j * xk1 * -x))
			else:
				ans = complex(-(xA * e**(1j * xk1 * -x) + xB * e**(-1j * xk1 * -x)))
		return choose_complex_portion(ans, fdata)
		if xeven:
			return ans.real
		else:
			return ans.imag

	def Run(self):
		try:
			self.think()
		except:
			(exc_type, exc_value, exc_traceback) = sys.exc_info()
			tb = traceback.format_tb(exc_traceback, 20)
			del exc_traceback
			self.Log('\nUnhandled %s exception\n'%(str(exc_type)), -1)
			self.Log('Message: %s\n'%(str(exc_value)), -1)
			for event in tb:
				for line in event.split('\n'):
					self.Log('\t'+line+'\n', -1)
			self.Log('Error performing computation, described above.\nAre your parameters outrageous?\n', -1)

		LogFunc('Done.\n', 2)
		self.NotifyOfComplete()

	def Stop(self):
		self.stop_now = True

	def think(self):
		global a, w, m, V, E, A, B, C, D, epsilon
		global lastE, even, case, even_incr, odd_incr, even_x, odd_x, limit, start_n
		a = self.a
		w = self.w
		m = self.m
		V = self.v
		if V == 0 or w == 0:
			V = 0
			w = 0
		self.Log('Solving a = %.15g w = %.15g m = %.15g V = %.15g\n'%(a, w, m, V))
		if not self.cont:
			even = EVEN
			case = E_LT_V
			even_incr = eV
			odd_incr = eV
			even_x = odd_x = 0
			limit = V
			start_n = 0
			lastE = None
		for i in range(start_n,self.n):
			if self.stop_now: break
			self.Log('E%d ='%(i+1), 3)
			sys.stdout.flush()
			while True: # will happen at most twice
				lfunc = None
				if case == E_LT_V:
					lfunc = lambda setE: (det_setE(matrices[case][even], setE)).imag
				else:
					lfunc = lambda setE: (det_setE(matrices[case][even], setE)).real
				if even == EVEN:
					(E, even_x, even_incr) = solve(lfunc, even_x, even_incr, limit)
				else:
					(E, odd_x, odd_incr) = solve(lfunc, odd_x, odd_incr, limit)
				if E == None:
					limit = None
					case = E_GT_V
					odd_x = even_x = V
					odd_incr = even_incr = eV
					self.Log('---', 3)
					continue
				break
			self.Log('\t%.15g eV'%(E/eV), 3)
			if lastE > E:
				self.Log("ARGHHHHH %.15g > %.15g\n"%(lastE/eV, E/eV), -1)
			if lastE != None:
				self.Log("\t(dE = %g)\n"%((E-lastE)/eV), 1)
			else:
				self.Log("\n", 3)

			# Find the nullspace of the matrix.
			M = []
			for row in matrices[case][even]:
				r = []
				for col in row:
					r.append(col())
				M.append(r)
			epsilon = E/1e5
			self.Log('\tfinding nullspace...\n')
			[A, B, C, D] = find_nullspace(M)
			#if str(A) != str(B.conjugate()):
				#self.Log('OOO: a non-conjugate (E%d, A=%s, B=%s)\n'%(i+1,str(A),str(B)), 1)
			if case == E_LT_V:
				efunc = self.func_e_lt_v
			else:
				efunc = self.func_e_gt_v

			norm_pass = 1
			norm_coeff = 0
			fdata = [a, w, V, m, A, B, C, D, E, even == EVEN]
			while (abs(norm_coeff) > 1e6 or abs(norm_coeff) < 1e-6):
				self.Log('\tnormalising (pass %d)...'%norm_pass)
				norm_coeff = sqrt(2 * (integrate(
					lambda x, fdata: (efunc(x, fdata)**2),
					[a, w, V, m, A, B, C, D, E, even == EVEN], -a, 0, 1000)))
				self.Log(' coeff=%s\n'%(str(norm_coeff)))
				[A, B, C, D] = map((lambda x: x/norm_coeff), [A, B, C, D])
				norm_pass += 1

			fdata = [a, w, V, m, A, B, C, D, E, even == EVEN]

			if even == EVEN:
				# Orient the even cases upwards.
				if efunc(0, fdata) < 0:
					[A, B, C, D] = map((lambda x: -x), [A, B, C, D])
			else:
				# Orient the odd cases with a -ve derivate at 0.
				if efunc(-a/1000000.0, fdata) < 0:
					[A, B, C, D] = map((lambda x: -x), [A, B, C, D])

			fdata = [a, w, V, m, A, B, C, D, E, even == EVEN]
			self.Log('\tA=%s\n\tB=%s\n\tC=%s\n\tD=%s\n'%(str(A), str(B), str(C), str(D)))
			self.NotifyOfE(i+1, E, efunc, fdata)

			lastE = E

			# And now switch even/odd
			even ^= 1
			start_n = i+1
